#
# Copyright (C) 2007 Saket Sathe
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#
# $LastChangedBy: xaeroman $
# $LastChangedDate: 2007-07-24 19:49:16 +0530 (Tue, 24 Jul 2007) $
# $LastChangedRevision: 30 $
# 
#

from BaseElements import *
from PackageExceptions import *

""" A module for representing graphs

    @namespace Graph
"""

class NumberedEdgeGraph:

    """ Represents a numbered edge graph. Numbered edges are required to distinguish multiple edges between
        same set of vertices. This class also provides an indexed vertex and edge sets. These indices have certain advantages
        while computing in-degree and out-degree distributions.
        
    """
    
    def __init__(self):
        """ Constructs a numbered edge graph
        """
        ## Dictionary of edges, indexed by edge number
        self.edgeIndex = {}

        ## Dictionary of vertices, indexed by vertex number
        self.vertexIndex = {}

        ## Last edge number assigned
        self.__lastEdgeNumber = 0

        ## Dictionary of in-degree counts. Used for efficiently computing in-degree distribution
        self.__inDegreeCount = {}

        ## Dictionary of out-degree counts. Used for efficiently computing out-degree distribution
        self.__outDegreeCount = {}

    def addEdge(self, edge):
        """ Adds an edge to a graph. It also updates the vertex and edge indices. 

            @param edge Edge of type BaseElements::Edge to be added to the graph
        """
        self.edgeIndex[self.__lastEdgeNumber] = edge
        self.__lastEdgeNumber += 1

        startVertex = edge.startVertex
        endVertex = edge.endVertex

        startVertexNumber = startVertex.vertexNumber
        endVertexNumber = endVertex.vertexNumber

        vertexIndex = self.vertexIndex

        if startVertexNumber not in vertexIndex:
            vertexIndex[startVertexNumber] = startVertex

        if endVertexNumber not in vertexIndex:
            vertexIndex[endVertexNumber] = endVertex
            
        try:
            self.__outDegreeCount[startVertexNumber] += 1
        except KeyError:
            self.__outDegreeCount[startVertexNumber] = 1
            
        try:
            self.__inDegreeCount[endVertexNumber] += 1
        except KeyError:
            self.__inDegreeCount[endVertexNumber] = 1

    def deleteEdge(self, edgeNumber):
        """ Delete an edge

            @param edgeNumber Edge number to be deleted
        """
        del self.edgeIndex[edgeNumber]

    def addVertex(self, vertexNumber):
        """ Adds a vertex. Should be used with care

            @param vertexNumber Vertex number of vertex to be added
            @throws PackageExceptions::VertexError
        """
        try:
            self.vertexIndex[vertexNumber]
            raise VertexError(vertexNumber, ErrorMessages.vertexAlreadyExists)
        except KeyError:
            self.vertexIndex[vertexNumber] = Vertex(vertexNumber)
            return 

    def deleteVertex(self, vertexNumber):
        """ Deletes a vertex. Should be used with care

            @param vertexNumber Vertex number to be deleted
        """
        del self.vertexIndex[vertexNumber]

    def getEdges(self):
        """ Get all graph edges

            @return edgeIndex Dictionary of edges, indexed by edge number
        """
        return self.edgeIndex

    def getVertices(self):
        """ Get all graph vertices

            @return vertexIndex Dictionary of vertices, indexed by vertex number
        """
        return self.vertexIndex
    
    def getOutNeighbors(self, vertexNumber):
        """ Get out-neighbors for a vertex

            @param vertexNumber Vertex number for which out-neighbors have to be obtained
            @return outNeighbors List of out-neighbors. Each element of type BaseElements::Vertex
        """
        outNeighbors = []
        edges = self.edgeIndex.values()
        for edge in edges:
            if edge.startVertex.vertexNumber == vertexNumber:
                outNeighbors.append(edge.startVertex)
        return outNeighbors

    def getInNeighbors(self, vertexNumber):
        """ Get in-neighbors for a vertex

            @param vertexNumber Vertex number for which in-neighbors have to be obtained
            @return inNeighbors List of in-neighbors. Each element of type BaseElements::Vertex
        """
        inNeighbors = []
        edges = self.edgeIndex.values()
        for edge in edges:
            if edge.endVertex.vertexNumber == vertexNumber:
                inNeighbors.append(endVertex)
        return inNeighbors

    def getNumberOfOutNeighbors(self, vertexNumber):
        """ Get number of out-neighbors for a vertex

            @param vertexNumber Vertex number for which number of out-neighbors have to be obtained
            @return Number of out-neighbors
        """ 

        return self.__outDegreeCount[vertexNumber]

    def getNumberOfInNeighbors(self, vertexNumber):
        """ Get number of in-neighbors for a vertex

            @param vertexNumber Vertex number for which number of in-neighbors have to be obtained
            @return Number of in-neighbors
        """ 
        return self.__inDegreeCount[vertexNumber]

    def getInDegreeDistribution(self):
        """ Get in-degree distribution

            @return inDegreeDistribution Dictionary indexed on in-degree. Values are the number of nodes for a in-degree
        """
        inDegreeDistribution = {}
        inDegreeCount = self.__inDegreeCount
        getNumberOfInNeighbors = self.getNumberOfInNeighbors
        vertexNumbers = inDegreeCount.keys()
        for vertexNumber in vertexNumbers:
            numberOfInNeighbors = inDegreeCount[vertexNumber]
            try:
                inDegreeDistribution[numberOfInNeighbors] += 1
            except KeyError:
                inDegreeDistribution[numberOfInNeighbors] = 1
        return inDegreeDistribution

    def getOutDegreeDistribution(self):
        """ Get out-degree distribution

            @return outDegreeDistribution Dictionary indexed on in-degree. Values are the number of nodes for a out-degree
        """
        outDegreeDistribution = {}
        outDegreeCount = self.__outDegreeCount
        getNumberOfOutNeighbors = self.getNumberOfOutNeighbors
        vertexNumbers = outDegreeCount.keys()
        for vertexNumber in vertexNumbers:
            numberOfOutNeighbors = outDegreeCount[vertexNumber]
            try:
                outDegreeDistribution[numberOfOutNeighbors] += 1
            except KeyError:
                outDegreeDistribution[numberOfOutNeighbors] = 1
        return outDegreeDistribution

    def getVerticesByInDegree(self, degree):
        """ Gets all the vertices with a particular in-degree

            @param degree In-degree to look for
            @return degreeNodes List of vertices. Each element of type BaseElements::Vertex
        """
        degreeNodes = []
        vertexIndex = self.vertexIndex
        inDegreeCount = self.__inDegreeCount
        for vertexNumber in inDegreeCount:
            if inDegreeCount[vertexNumber] == degree:
                degreeNodes.append(vertexIndex[vertexNumber])
        return degreeNodes

    def getVerticesByOutDegree(self, degree):
        """ Gets all the vertices with a particular out-degree

            @param degree Out-degree to look for
            @return degreeNodes List of vertices. Each element of type BaseElements::Vertex
        """
        degreeNodes = []
        vertexIndex = self.vertexIndex
        outDegreeCount = self.__inDegreeCount
        for vertexNumber in outDegreeCount:
            if outDegreeCount[vertexNumber] == degree:
                degreeNodes.append(vertexIndex[vertexNumber])
        return degreeNodes
    
    def writeEdges(self, fileName, format):
        """ Write edges to file

            @param fileName File name to store edges in
            @param format Format of output file. Can take values: <br>
                          'simple' = simple format <br>
                          'dot' = format compatible with 'dot' command
                          
        """
        edges = self.edgeIndex.values()
        if format == 'simple':
            f = open(fileName,'w')
            for edge in edges:
                f.write("%s -> %s\n" % (edge.startVertex.vertexNumber, edge.endVertex.vertexNumber))
            f.close()
            
        elif format == 'dot':
            f = open(fileName,'w')
            f.write("digraph G { \n")
            for edge in edges:
                f.write("%s -> %s;\n" % (edge.startVertex.vertexNumber, edge.endVertex.vertexNumber))
            f.write("} \n")
            f.close()

    def readEdges(self, fileName, format):
        """ Read edges from file

            @param fileName File name to read edges from
            @param format Format of input file. Can take values: <br>
                          'simple' = simple format
        """
        f = open(fileName)
        if format == 'simple':
            edgesRaw = f.read().split("\n")

            if edgesRaw[-1] == '': edgesRaw = edgesRaw[:-1]

            for edge in edgesRaw:
                [startVertex, endVertex] = edge.split("->")
                newEdge = Edge(Vertex(int(startVertex)), Vertex(int(endVertex)))
                self.addEdge(newEdge)

    def findEdge(self, edgeNumber):
        """ Find edge with a given edge number

            @param edgeNumber Edge number to look for
            @throws PackageExceptions::EdgeError
            @return Matched edge of type BaseElements::Edge
            
        """
        try:
            return self.edgeIndex[edgeNumber]
        except KeyError:
            raise EdgeError(edgeNumber, ErrorMessages.edgeNotFound)
    
    def findVertex(self, vertexNumber):
        """ Find vertex with a given vertex number

            @param vertexNumber Vertex number to look for
            @throws PackageExceptions::VertexError
            @return Matched vertex of type BaseElements::Vertex
        """
        try:
            return self.vertexIndex[vertexNumber]
        except KeyError:
            raise VertexError(vertexNumber, ErrorMessages.vertexNotFound)

    def hasVertex(self, vertexNumber):
        """ Checks if vertex is present

            @param vertexNumber Vertex number of the vertex to check
            @return 0 if found. 1 if not found
        """
        try:
            rs = self.findVertex(vertexNumber)
            return 0
        except VertexError, e:
            return 1

